#ifndef __mOCS_H__
#define __mOCS_H__

/*! 
    \file mOCS.h
    \brief 3D Orthogonal Coordinate System
    \author Douglas H. Cox
*/

//============================================================================
//    REQUIRED HEADERS
//============================================================================
#include "mMath.h"
#include "mVec3.h"
#include "mFrame.h"

//============================================================================
//    DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS / CLASS FORWARDS
//============================================================================
class mQuat;
class mMatrix;

//============================================================================
//    FUNCTION PROTOTYPES
//============================================================================

//============================================================================
//    CLASS PROTOTYPES
//============================================================================

//! 3D Orthogonal Coordinate System
/*! mOCS represents a 3D Orthogonal Coordinate System.
    \sa http://www.loonygames.com/content/1.29/cotc/
*/
class MMATH_DLL mOCS
{
private:
  mVec3  o; //!< OCS origin.
  mFrame f; //!< OCS axial frame.
  mVec3  s; //!< OCS axis scale.


public:
  enum TYPE
  {
    IDENTITY = 0x00,
    ORIGIN   = 0x01,
    FRAME    = 0x02,
    SCALE    = 0x04
  };
  
private:
  unsigned type; //!< Specifies the valid components of the OCS.

  friend class mVec3;
  friend class mRay;

public:
  //! Default constructor
  /*! Initialized to origin at zero, cartesian axes, and unit scale. */
  mOCS()
  {
    type = IDENTITY;
    makeIdentity();
  }

  //! Copy constructor
  mOCS(const mOCS& src);

  //! Constructor specifying origin, axial frame, and scale.
  /*! Example:
      
      mOCS ocs(&origin, &rot, NULL);

      would create an OCS with an \em origin and a \em frame component.  
      Types not specified by \em type are set to an identity value. */
  mOCS(const mVec3& origin, 
       const mFrame& frame, 
       const mVec3& scale,
       unsigned type = ORIGIN | FRAME | SCALE);

  //! Origin at zero, cartesian axes, unit scale.
  void makeIdentity();

  //! Creates an ocs with specified origin, unit scale, and cartesian axes.
  void makeOrigin(const mVec3& origin)
  { 
    makeOrigin(origin[0], origin[1], origin[2]); 
  }

  //! Creates an ocs with specified origin, unit scale, and cartesian axes.
  void makeOrigin(float x, float y, float z);


  //! Creates an ocs with specified frame, origin at zero, and unit scale.
  void makeFrame(const mFrame& frame);

  //! Creates an ocs with specified scale, origin at zero, and cartesian axes.
  void makeScale(const mVec3& scale)
  {
    makeScale(scale[0], scale[1], scale[2]);
  }

  //! Creates an ocs with specified scale, origin at zero, and cartesian axes.
  void makeScale(float x, float y, float z);

  //! Assignment
  void set(const mOCS& src);

  //! Assignment specifying origin, axial frame, and scale.
  /*! Types not specified by \em type are set to an identity value. */
  void set(const mVec3& origin, 
           const mFrame& frame, 
           const mVec3& scale,
           unsigned type = ORIGIN | FRAME | SCALE);
  
  //! Assignment 
  void setOrigin(const mVec3& origin)
  {
    o.set(origin);
    type |= ORIGIN;
  }

  //! Assignment 
  void setOrigin(float x, float y, float z)
  {
    o.set(x, y, z);
    type |= ORIGIN;
  }

  //! Assignment 
  void setFrame(const mFrame& frame)
  {
    f.set(frame);
    type |= FRAME;
  }

  //! Sets the axial frame to the rotation defined by the quaternion q.
  /*! This call only affects the axial frame. */
  void setQuat(const mQuat& q)
  { 
    f.setQuat(q);
    type |= FRAME;
  }

  //! Sets the axial frame to a rotation about a unit vector n. a is in degrees.
  /*! This call only affects the axial frame. */
  void setAxisAngle(const mVec3& n, float a)
  {
    f.setAxisAngle(n, a);
    type |= FRAME;
  }
 
  //! Sets this to a rotation defined by the heading and pitch in degrees.
  /*! This assumes a roll of 0.0f and is faster than the full setHPR method. */
  void setHP(float heading, float pitch)
  {
    f.setHP(heading, pitch);
    type |= FRAME;
  }

  //! Sets this to a rotation defined by heading, pitch, and roll in degrees.
  void setHPR(float heading, float pitch, float roll)
  {
    f.setHPR(heading, pitch, roll);
    type |= FRAME;
  }

  //! Sets this to a rotation defined by heading, pitch, and roll in degrees.
  void setHPR(const mVec3& hpr)
  {
    setHPR(hpr.v[0], hpr.v[1], hpr.v[2]);
  }

  //! Assignment 
  void setScale(const mVec3& scale)
  {
    s.set(scale);
    type |= SCALE;
  }

  //! Assignment 
  void setScale(float x, float y, float z)
  {
    s.set(x, y, z);
    type |= SCALE;
  }

  //! Assignment
  mOCS& operator = (const mOCS& src) 
  { 
    set(src); 
    return *this; 
  }

  //! Accessor
  void get(mVec3& origin, mFrame& frame, mVec3& scale) const;  
 
  //! Accessor
  const mVec3* getOrigin() const
  {
    return &o;    
  }

  //! Accessor
  const mFrame* getFrame() const
  {
    return &f;
  }

  //! Accessor
  const mVec3* getScale() const
  {
    return &s;
  }

  //! Returns the type of valid components in the OCS.
  unsigned getType() const
  {
    return type;
  }

  //! Returns true if the two OCSs are equal.
  bool operator == (const mOCS& ocs) const;

  //! Returns true if the two OCSs are not equal.
  bool operator != (const mOCS& ocs) const
  {
    return !(operator==(ocs));
  }

  //! Returns true is this is within some tolerance value of \em ocs.
  bool approxEqual(const mOCS& ocs, float tol = mZeroTol) const;

  //! Computes the inverse of f. 
  void inverse(const mOCS& ocs);

  //! Transforms ocs into 'to's coordinate system.
  /*! \param ocs OCS that will be transformed
      \param to OCS that ocs will be transformed to */
  void xformTo(const mOCS& ocs, const mOCS& to);

  //! Transforms ocs from 'from's coordinate system.
  /*! \param ocs OCS that will be transformed
      \param from OCS that ocs will be transformed from */
  void xformFrom(const mOCS& ocs, const mOCS& from);
 

  //! Builds a matrix that transforms object into this OCS.
  /*! This builds an OpenGL compatible matrix. That is,
      one that can be loaded into a modelview matrix. */
  void getToMatrix(mMatrix& matrix) const;
  
  //! Builds a matrix that transforms object from this OCS.
  /*! This builds an OpenGL compatible matrix. That is,
      one that can be loaded into a modelview matrix. */
  void getFromMatrix(mMatrix& matrix) const;
};

//****************************************************************************
//**    END HEADER mOCS.h
//****************************************************************************
#endif // __mOCS_H__
