#ifndef __mQuat_H__
#define __mQuat_H__

/*! 
    \file mQuat.h
    \brief Quaterion
    \author Douglas H. Cox
*/

//============================================================================
//    REQUIRED HEADERS
//============================================================================
#include "mMath.h"
#include "mVec4.h"
#include "mVec3.h"
#include "mFrame.h"

//============================================================================
//    DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS / CLASS FORWARDS
//============================================================================


//============================================================================
//    FUNCTION PROTOTYPES
//============================================================================

//============================================================================
//    CLASS PROTOTYPES
//============================================================================

//! Quaternion
/*! mQuat represents a quaternion. 
    \sa http://mathworld.wolfram.com/Quaternion.html
*/
class MMATH_DLL mQuat : public mVec4
{
public:
  //! Default constructor
  /*! No initialization is done here. */
  mQuat() {}

  //! Copy constructor
  mQuat(const mQuat& src)
  { 
    v[0] = src.v[0]; 
    v[1] = src.v[1]; 
    v[2] = src.v[2];
    v[3] = src.v[3];
  }

  //! Constructor from 4 float values
  mQuat(float x, float y, float z, float w)
  {
    v[0] = x; 
    v[1] = y; 
    v[2] = z;
    v[3] = w;
  }

  //! Constructor from an array of 4 float values
  mQuat(float *src)
  {
    v[0] = src[0];
    v[1] = src[1];
    v[2] = src[2];
    v[3] = src[3];
  }

  //! Sets this to its conjugate
  void conjugate()
  {
    v[0] = -v[0];
    v[1] = -v[1];
    v[2] = -v[2];
  }

  //! Set this to the conjugate of q
  void conjugate(const mQuat& q)
  {
    v[0] = -q.v[0];
    v[1] = -q.v[1];
    v[2] = -q.v[2];
    v[3] = q.v[3];
  }

  //! Computes the inverse of this. this = 1 / this
  void inverse();

  //! Computes the inverse of q. this = 1 / q
  void inverse(const mQuat& q);

  //! Multiply: this *= q
  void mult(const mQuat& q)
  {
    mult(*this, q);
  }

  //! Multiply: this = q1 * q2
  void mult(const mQuat& q1, const mQuat& q2);

  //! Spherical linear interpolation between q1 and q2.
  void slerp(const mQuat& q1, const mQuat& q2, float t);

  //! Sets this to the rotation about a unit vector n. a is in degrees.
  void setAxisAngle(const mVec3& n, float a);

  //! Gets an angle about a unit vector n defined by this quaternion.
  void getAxisAngle(mVec3& n, float& a) const;

  //! Sets this to the rotation defined by the heading and pitch in degrees.
  /*! This assumes a roll of 0.0f and is faster than the full setHPR method. */
  void setHP(float heading, float pitch);

  //! Returns the heading and pitch defined by this quaternion.
  void getHP(float &heading, float &pitch) const;

  //! Sets this to the rotation defined by heading, pitch, and roll in degrees.
  void setHPR(float heading, float pitch, float roll);

  //! Sets this to the rotation defined by heading, pitch, and roll in degrees.
  void setHPR(const mVec3& hpr)
  {
    setHPR(hpr.v[0], hpr.v[1], hpr.v[2]);
  }

  //! Returns the heading and pitch defined by this quaternion.
  /*! \todo Implement this. */
  void getHPR(mVec3& hpr) const;

  //! Set this to the rotation defined by frame.
  void setFrame(const mFrame& frame)
  {
    frame.getQuat(*this); 
  }

  //! Assigns a frame to the rotation defined by this quaternion.
  void getFrame(mFrame& frame) const
  {
    frame.setQuat(*this);
  }

};

//****************************************************************************
//**    END HEADER mQuat.h
//****************************************************************************
#endif // __mQuat_H__
